try
{
  if (QMessageBox.question(mywindow, qsTr("Backed Up?"),
                           qsTr("Have you backed up your database? You really should "
                              + "back up your database before running this utility."),
                           QMessageBox.Yes | QMessageBox.No) == QMessageBox.No)
    mywindow.close()

  var _addCountry = mywindow.findChild("_addCountry");
  var _close      = mywindow.findChild("_close");
  var _edit       = mywindow.findChild("_edit");
  var _finish     = mywindow.findChild("_finish");
  var _fixBulk    = mywindow.findChild("_fixBulk");
  var _list       = mywindow.findChild("_list");
  var _query      = mywindow.findChild("_query");

  var _params = {};

  var map = { addr:     { id: 'addr_id',     mode: 'edit',      window: 'address'             },
              quhead:   { id: 'quhead_id',   mode: 'editQuote', window: 'salesOrder'          },
              cohead:   { id: 'sohead_id',   mode: 'edit',      window: 'salesOrder'          },
              invchead: { id: 'invchead_id', mode: 'edit',      window: 'invoice'             },
              cmhead:   { id: 'cmhead_id',   mode: 'edit',      window: 'creditMemo'          },
              pohead:   { id: 'pohead_id',   mode: 'edit',      window: 'purchaseOrder'       },
              cohist:   { id: 'sohist_id',   mode: 'edit',      window: 'salesHistoryInformation' },
              ccard:    { id: 'ccard_id',    mode: 'edit',      window: 'creditCard'          },
              rahead:   { id: 'rahead_id',   mode: 'edit',      window: 'returnAuthorization' },
              tohead:   { id: 'tohead_id',   mode: 'edit',      window: 'transferOrder'       }
            },
       exceptionmsg = qsTr("fixCountries exception @ %1: %2");

  _list.addColumn(qsTr("Record Type"), -1, Qt.AlignLeft, true, "type");
  _list.addColumn(qsTr("Details"),     -1, Qt.AlignLeft, true, "detail");
  _list.addColumn(qsTr("Number"),      -1, Qt.AlignLeft, true, "number");
  _list.addColumn(qsTr("Country"),     -1, Qt.AlignLeft, true, "country");

  _addCountry.clicked.connect(sAddCountry);
  _close.clicked.connect(mywindow, "close");
  _edit.clicked.connect(sEdit);
  _finish.clicked.connect(sFinish);
  _fixBulk.clicked.connect(sFixBulk);
  _list.valid.connect(sHandleButtons);
  _list['itemDoubleClicked(XTreeWidgetItem*,int)'].connect(sEdit);
  _query.clicked.connect(sFillList);

}
catch (e)
{
  QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
}

function set(params)
{
  try
  {
    if ("cohist" in params)
      _params.cohist = true;
    if ("asohist" in params)
      _params.asohist = true;
    if ("rahead" in params)
      _params.rahead = true;
    if ("tohead" in params)
      _params.tohead = true;
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sFillList()
{
  try
  {
    var params = {};
    if ("cohist" in _params)
      params.cohist = true;
    if ("asohist" in _params)
      params.asohist = true;
    if ("rahead" in _params)
      params.rahead = true;
    if ("tohead" in _params)
      params.tohead = true;
    if (mainwindow.key)
      params.key = mainwindow.key;
    params.address = qsTr("Address");
    params.quote = qsTr("Quote");
    params.salesorder = qsTr("Sales Order");
    params.invoice = qsTr("Invoice");
    params.salescredit = qsTr("Sales Credit");
    params.purchaseorder = qsTr("Purchase Order");
    params.saleshistory = qsTr("Sales History");
    params.archivedsaleshistory = qsTr("Archived Sales History");
    params.creditcard = qsTr("Credit Card");
    params.returnauthorization = qsTr("Return Authorization");
    params.transferorder = qsTr("Transfer Order");
    params.configuration = qsTr("Configuration");
    params.billto = qsTr("Bill-To");
    params.shipto = qsTr("Ship-To");
    params.vendor = qsTr("Vendor");
    params.source = qsTr("Source");
    params.destination = qsTr("Destination");
    params.addressdefault = qsTr("Address Default");
    params.remitto = qsTr("Remit-To");

    var qry = toolbox.executeDbQuery('fixcountry', 'strictcountrycheck', params);
    if (qry.lastError().type != QSqlError.NoError)
      throw new Error(qry.lastError().text);

    _list.populate(qry);
    _finish.enabled = privileges.check("#superuser") && mainwindow.key && (qry.size() == 0);
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sHandleButtons(valid)
{
  try
  {
    var key     = valid ? _list.currentItem().rawValue("type") : '',
        canEdit = (key != "asohist");
    _edit.enabled = valid && canEdit;
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sFixBulk()
{
  try
  {
    toolbox.openWindow("bulkFixCountry", mywindow, Qt.WindowModal, Qt.Dialog);
    toolbox.lastWindow().set(_params);
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sEdit()
{
  try
  {
    var key = _list.currentItem().rawValue('type'),
        detail = _list.currentItem().rawValue('detail'),
        newdlg, params = {};

    if (key in map)
    {
      params[map[key].id] = _list.id();
      params.mode = map[key].mode || 'edit';
      newdlg = toolbox.openWindow(map[key].window, mywindow, Qt.Dialog, Qt.WindowModal);
      newdlg.set(params);
      if (newdlg && 'exec' in newdlg && newdlg.exec() != QDialog.Rejected)
        sFillList();
    }
    else if (key == "metric")
    {
      if (detail == "addressdefault")
        params.module = Xt.CRMModule;
      else if (detail == "remitto")
        params.module = Xt.AccountingModule;
      newdlg = toolbox.openWindow('setup', mywindow, Qt.Dialog, Qt.WindowModal);
      newdlg.set(params);

      if (detail == "remitto")
        newdlg.findChild("_tab").setCurrentWidget(newdlg.findChild("_ar"));

      if (newdlg.exec() != QDialog.Rejected)
        sFillList();
    }
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sAddCountry()
{
  try
  {
    var newdlg = toolbox.openWindow("country", mywindow, Qt.Dialog, Qt.WindowModal);
    toolbox.lastWindow().set({ mode: "new" });
    if (newdlg.exec() != QDialog.Rejected)
      sFillList();
  }
  catch (e)
  {
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}

function sFinish()
{
  try
  {
    var columns = ["addr_country",
                   "quhead_billtocountry", "quhead_shiptocountry",
                   "cohead_billtocountry", "cohead_shiptocountry",
                   "invchead_billto_country", "invchead_shipto_country",
                   "cmhead_billtocountry", "cmhead_shipto_country",
                   "pohead_vendcountry", "pohead_shiptocountry",
                   "ccard_country", "metric_value"];

    if ("cohist" in _params)
      columns = columns.concat(["cohist_billtocountry", "cohist_shiptocountry"]);
    if ("asohist" in _params)
      columns = columns.concat(["asohist_billtocountry", "asohist_shiptocountry"]);
    if ("rahead" in _params)
      columns = columns.concat(["rahead_billtocountry", "rahead_shipto_country"]);
    if ("tohead" in _params)
      columns = columns.concat(["tohead_srccountry", "tohead_destcountry"]);

    var params = {};
    if (mainwindow.key)
      params.key = mainwindow.key;

    toolbox.executeQuery("BEGIN;");

    var qry = toolbox.executeQuery("UPDATE country "
                                 + "   SET country_abbr = 'GB' "
                                 + " WHERE country_abbr = 'UK';");

    if (qry.lastError().type != QSqlError.NoError)
      throw new Error(qry.lastError().text);

    for (i = 0; i < columns.length; i++)
    {
      params.table = columns[i].split("_")[0];
      params.column = columns[i];

      if (columns[i] == "metric_value")
        params.metric = true;
      else if (params.metric)
        delete params.metric;

      qry = toolbox.executeQuery("ALTER TABLE <? literal('table') ?> DISABLE TRIGGER ALL;", params);

      if (qry.lastError().type != QSqlError.NoError)
        throw new Error(qry.lastError().text);

      if (columns[i] == "ccard_country")
      {
        if (mainwindow.key)
          qry = toolbox.executeQuery("UPDATE ccard "
                                   + "   SET ccard_country = encrypt(setbytea(country_abbr), "
                                   + "                               setbytea(<? value('key') ?>), "
                                   + "                               'bf') "
                                   + "  FROM country "
                                   + " WHERE lower(country_name) = "
                                   + " lower(formatbytea(decrypt(setbytea(ccard_country), "
                                   + "                           setbytea(<? value('key') ?>), "
                                   + "                           'bf')));",
                                     params);
        else
          QMessageBox.critical(mywindow, qsTr("Cannot update ccard table"),
                               qsTr("Cannot update the ccard table without having the encryption key"));
      }
      else
      {
        qry = toolbox.executeQuery("UPDATE <? literal('table') ?> "
                                 + "   SET <? literal('column') ?> = country_abbr "
                                 + "  FROM country "
                                 + " WHERE lower(country_name) = lower(<? literal('column') ?>) "
                                 + "<? if exists('metric') ?> "
                                 + "   AND metric_name IN ('DefaultAddressCountry', "
                                 + "                       'remitto_country') "
                                 + "<? endif ?>;",
                                   params);
      }

      if (qry.lastError().type != QSqlError.NoError)
        throw new Error(qry.lastError().text);

      qry = toolbox.executeQuery("ALTER TABLE <? literal('table') ?> ENABLE TRIGGER ALL;", params);

      if (qry.lastError().type != QSqlError.NoError)
        throw new Error(qry.lastError().text);
    }

    var params = { finished: true };
    if (mainwindow.key)
      params.key = mainwindow.key;
    if ("cohist" in _params)
      params.cohist = true;
    if ("asohist" in _params)
      params.asohist = true;
    if ("rahead" in _params)
      params.rahead = true;
    if ("tohead" in _params)
      params.tohead = true;

    qry = toolbox.executeDbQuery('fixcountry', 'strictcountrycheck', params);
    var remaining = [];
    while (qry.next())
      remaining.push(qsTr("<li>%1-%2-%3-%4</li>").arg(qry.value("type"))
                                                 .arg(qry.value("detail"))
                                                 .arg(qry.value("number"))
                                                 .arg(qry.value("country")));

    if (remaining.length > 0)
      throw new Error(qsTr("It looks as though %1 record(s) were not successfully "
                         + "migrated. Please requery and try again. If you have repeated "
                         + "problems contact xTuple Support.<ul>%2</ul>")
                         .arg(remaining.length).arg(remaining.join("")));
    else if (qry.lastError().type != QSqlError.NoError)
      throw new Error(qry.lastError().text);

    metrics.set("ISOCountries", true);

    toolbox.executeQuery("COMMIT;");

    QMessageBox.information(mywindow, qsTr("Successfully Migrated"),
                            qsTr("Successfully migrated country data. You may now upgrade to "
                               + "version 5.0.0 or later."));
  }
  catch (e)
  {
    toolbox.executeQuery("ROLLBACK;");
    QMessageBox.critical(mywindow, "fixCountries", exceptionmsg.arg(e.lineNumber).arg(e.message));
  }
}
